'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.mainVisitor = exports.processMacros = exports.collectMacros = exports.collectMacros2 = exports.collectMacros1 = undefined;

var _symbols = require('./symbols');

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _babelTraverse = require('babel-traverse');

var _babelTraverse2 = _interopRequireDefault(_babelTraverse);

var _Macro = require('./Macro');

var _Macro2 = _interopRequireDefault(_Macro);

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

var collectMacros1 = exports.collectMacros1 = {
  CallExpression: function CallExpression(path) {
    "use strict";

    _processMacro(path, true);
  }
};

var collectMacros2 = exports.collectMacros2 = {
  LabeledStatement: {
    exit: function exit(path) {
      "use strict";

      if (path.node.label.name === 'macro') {
        path.remove();
      }
    }
  },
  VariableDeclarator: function VariableDeclarator(path) {
    "use strict";

    var parent1 = path.parentPath.parentPath,
        parent2 = parent1.parentPath;
    if (t.isLabeledStatement(parent1) && parent1.node.label.name === 'macro' || t.isLabeledStatement(parent2) && parent2.node.label.name === 'macro') {
      _Macro2.default.register(path.node.id.name, path.get('init'), parent2.scope);
      path.remove();
    }
  },
  FunctionDeclaration: function FunctionDeclaration(path) {
    "use strict";

    var node = path.node,
        parent1 = path.parentPath,
        parent2 = parent1.parentPath;
    if (t.isLabeledStatement(parent1) && parent1.node.label.name === 'macro' || t.isLabeledStatement(parent2) && parent2.node.label.name === 'macro') {
      (0, _utils.warning)('FunctionDeclaration converted to macro. Incompatible with native JS behaviour' + (0, _utils.getLocationMessage)(node));
      _Macro2.default.register(node.id.name, path, parent2.scope);
      path.remove();
    }
  }
};

var collectMacros = exports.collectMacros = _babelTraverse2.default.visitors.merge([collectMacros1, collectMacros2]);

var processMacros = exports.processMacros = {
  CallExpression: function CallExpression(path) {
    "use strict";

    var node = path.node;

    if (node[_symbols.$processedByMacro]) {
      return;
    }
    _processMacro(path, false);
    node[_symbols.$processedByMacro] = true;
  }
};

/* @todo
 macros are deleted immediately after fin ding
 This prevents third-party plug-ins to make a transformation within their code
 Possible solutions:
 1. Before removing the macro execute other plugins on his body
 perhaps they are all available in the article
 2. remove the macros only when they are all applied
 3. temporary hack. search and process of macros only after all the work of other plug-ins

 last variant temporary is used
 */
var mainVisitor = exports.mainVisitor = {
  Program: {
    enter: function enter(path) {
      "use strict";

      path.traverse(collectMacros);
      path.traverse(processMacros);
    }
  }
};

function _processMacro(path, isBuiltinMacro) {
  "use strict";

  var node = path.node;

  if (t.isMemberExpression(node.callee)) {
    if (!node.callee.computed && _Macro2.default.getMacro(node.callee.object, path.scope, isBuiltinMacro) && _Macro2.default.getMacro(node.callee.property, path.scope, isBuiltinMacro)) {
      var head = node.callee.object;
      var tailId = node.callee.property;
      node.arguments.unshift(head);
      var macro = _Macro2.default.getMacro(tailId, path.scope, isBuiltinMacro);
      if (macro) {
        _Macro2.default.runMacro(path, macro, path.scope);
      }
    }
  } else {
    var macro = _Macro2.default.getMacro(node.callee, path.scope, isBuiltinMacro);
    if (macro) {
      _Macro2.default.runMacro(path, macro, path.scope);
    }
  }
}